@@ -5,7 +5,7 @@ import socket
 import socketserver
 import sys
 import threading
-from argparse import ArgumentParser
+from argparse import ArgumentParser, FileType
 from enum import IntEnum
 from logging.handlers import RotatingFileHandler
 from pathlib import Path
@@ -24,9 +24,8 @@ RANDOM = "Special:Random"
 NPREFETCH = 10
 HOST = "localhost"
 PORT = 48484
-SERVER_TIMEOUT = 1
 CLIENT_TIMEOUT = 3
-LOG_PATH = Path(__file__).parent / "proofaday.log"
+LOG_PATH = Path(__file__).parent
 
 
 class MsgKind(IntEnum):
@@ -45,43 +44,67 @@ class ProofHandler(socketserver.BaseRequestHandler):
         data, sock = self.request
         kind, msg = MsgKind(data[0]), data[1:].decode()
         server: ProofServer = self.server  # type: ignore[assignment]
+        logger = server.logger
+        logger.info("Received %s from (%s, %d)", kind, *self.client_address)
 
         if kind is MsgKind.CHECK:
             reply = ""
         elif kind is MsgKind.REQUEST:
+            logger.info("Fetching %s", msg)
             proof = server.fetch_proof(msg)
             reply = proof if proof is not None else ""
         elif kind is MsgKind.RANDOM:
+            logger.info("Dequeuing proof")
             reply = server.queue.get()
         elif kind is MsgKind.KILL:
+            logger.info("Shutting down server")
             sock.sendto(b"", self.client_address)
             server.shutdown()
+            logger.info("Server shutdown")
             return
         sock.sendto(reply.encode(), self.client_address)
 
 
 class ProofServer(socketserver.ThreadingUDPServer):
+    proof_timeout = 1
+    daemon_threads = True
+
     max_log_bytes = 1024 * 1024
     max_requests = 5
 
     def __init__(
-        self, port: int, limit: int, nprefetch: int, debug: bool, **kwargs: Any
+        self,
+        port: int,
+        limit: int,
+        nprefetch: int,
+        debug: int,
+        log_path: Path,
+        **kwargs: Any
     ) -> None:
         super().__init__((HOST, port), ProofHandler)
-        level = logging.DEBUG if debug else logging.INFO
-        self.logger = logging.getLogger(__name__)
-        self.logger.setLevel(level)
-        self.logger.addHandler(
-            RotatingFileHandler(
-                LOG_PATH,
+        level = {0: logging.NOTSET, 1: logging.INFO}.get(debug, logging.DEBUG)
+        self.logger = self.init_logger(level, log_path)
+        self.queue: Queue[str] = Queue(maxsize=nprefetch)
+        self.limit = limit
+        threading.Thread(
+            target=self.fetch_proofs, daemon=True, name="ServerLoop"
+        ).start()
+
+    def init_logger(self, level: int, path: Path) -> logging.Logger:
+        logger = logging.getLogger(__name__)
+        logger.setLevel(level)
+        if level != logging.NOTSET:
+            handler: logging.Handler = RotatingFileHandler(
+                path / "proofaday.log",
                 maxBytes=ProofServer.max_log_bytes,
                 backupCount=1,
                 encoding="utf8",
             )
-        )
-        self.queue: Queue[str] = Queue(maxsize=nprefetch)
-        self.limit = limit
-        threading.Thread(target=self.fetch_proofs, daemon=True).start()
+        else:
+            handler = logging.NullHandler()
+        handler.setFormatter(logging.Formatter("%(threadName)s: %(message)s"))
+        logger.addHandler(handler)
+        return logger
 
     def fetch_proof(self, name: Optional[str] = None) -> Optional[str]:
         if name is None:
@@ -89,7 +112,8 @@ class ProofServer(socketserver.ThreadingUDPServer):
         url = URL + name
 
         try:
-            html = BS(requests.get(url, timeout=SERVER_TIMEOUT).text, "html.parser")
+            data = requests.get(url, timeout=ProofServer.proof_timeout).text
+            html = BS(data, "html.parser")
             proof = Proof(html)
             self.logger.debug(repr(proof))
             return str(proof)
@@ -103,26 +127,29 @@ class ProofServer(socketserver.ThreadingUDPServer):
             )
         return None
 
-    def enqueue_proof(self) -> None:
-        proof = self.fetch_proof()
-        if proof is not None and (
-            self.limit is None or len(proof.split("\n")) <= self.limit
-        ):
+    def enqueue_proof(self, proof: str) -> None:
+        if self.limit is None or len(proof.split("\n")) <= self.limit:
             self.queue.put(proof)
 
     def fetch_proofs(self) -> NoReturn:
-        with futures.ThreadPoolExecutor(ProofServer.max_requests) as pool:
-            jobs: Set[futures.Future[None]] = set()
+        with futures.ThreadPoolExecutor(
+            max_workers=ProofServer.max_requests, thread_name_prefix="Fetcher"
+        ) as pool:
+            jobs: Set[futures.Future[Optional[str]]] = set()
             while True:
                 njobs = self.queue.maxsize - self.queue.qsize() - len(jobs)
                 if len(jobs) == 0:
                     njobs = max(njobs, 1)
-                jobs |= {pool.submit(self.enqueue_proof) for _ in range(njobs)}
-                done, _ = futures.wait(jobs, return_when=futures.FIRST_COMPLETED)
-                jobs -= done
+                jobs |= {pool.submit(self.fetch_proof) for _ in range(njobs)}
+                done, jobs = futures.wait(jobs, return_when=futures.FIRST_COMPLETED)
+
+                for job in done:
+                    proof = job.result()
+                    if proof is not None:
+                        self.enqueue_proof(proof)
 
 
-def daemon(*args: Any, **kwargs: Any) -> None:
+def daemon(*args: Any, **kwargs: Any) -> NoReturn:
     try:
         with ProofServer(*args, **kwargs) as server:
             server.serve_forever()
@@ -199,13 +226,15 @@ def main() -> None:
 
     parser = ArgumentParser(description="Fetch a random proof")
     parser.add_argument("name", nargs="?", default=None)
-    parser.add_argument("-d", "--debug", action="store_true")
+    parser.add_argument("-d", "--debug", action="count", default=0)
+    parser.add_argument("--log-path", dest="log_path", type=Path, default=LOG_PATH)
     parser.add_argument("-l", "--limit", type=pos, default=None)
     parser.add_argument(
         "-n", "--num-prefetch-proofs", dest="nprefetch", type=pos, default=NPREFETCH
     )
     parser.add_argument("-p", "--port", type=int, default=PORT)
     parser.add_argument("-t", "--timeout", type=float, default=CLIENT_TIMEOUT)
+    parser.add_argument("-o", "--output", type=FileType("w"), default=sys.stdout)
     parser.add_argument("-k", "--kill-server", dest="kill", action="store_true")
     args = parser.parse_args()
 
@@ -215,7 +244,7 @@ def main() -> None:
             sys.exit(not client.kill())
         if not client.check():
             client.spawn(**vars(args))
-        print(client.query(args.name))
+        print(client.query(args.name), file=args.output)
     except ProofServerError as e:
         sys.exit(str(e))
 
